@@ -3,10 +3,12 @@ from __future__ import annotations
 import abc
 import datetime
 
+from lime_uow import resources
+
 from lime_etl.domain import value_objects
 
 
-class TimestampAdapter(abc.ABC):
+class TimestampAdapter(resources.Resource[value_objects.Timestamp], abc.ABC):
     @abc.abstractmethod
     def now(self) -> value_objects.Timestamp:
         raise NotImplementedError
@@ -20,5 +22,17 @@ class TimestampAdapter(abc.ABC):
 
 
 class LocalTimestampAdapter(TimestampAdapter):
+    def close(self) -> None:
+        pass
+
+    def open(self) -> resources.Resource[value_objects.Timestamp]:
+        return self
+
+    def rollback(self) -> None:
+        pass
+
+    def save(self) -> None:
+        pass
+
     def now(self) -> value_objects.Timestamp:
         return value_objects.Timestamp(datetime.datetime.now())
