@@ -246,7 +246,7 @@ class SingleChar(ValueObject):
         super().__init__(value)
 
 
-class JobName(ValueObject):
+class _DbName(ValueObject):
     def __init__(self, value: str, /):
         if value is None:
             raise ValueError(
@@ -266,6 +266,16 @@ class JobName(ValueObject):
         super().__init__(value)
 
 
+class BatchName(_DbName):
+    def __init__(self, value: str, /):
+        super().__init__(value)
+
+
+class JobName(_DbName):
+    def __init__(self, value: str, /):
+        super().__init__(value)
+
+
 class ResourceName(_NonEmptyStr):
     ...
 
@@ -320,6 +330,11 @@ class DaysToKeep(_PositiveInt):
 class ExecutionMillis(_PositiveInt):
     ...
 
+    @staticmethod
+    def calculate(*, start_time: Timestamp, end_time: Timestamp) -> ExecutionMillis:
+        elapsed_millis = int((end_time.value - start_time.value).total_seconds() * 1000)
+        return ExecutionMillis(elapsed_millis)
+
 
 class TimeoutSeconds(_PositiveInt):
     ...
@@ -329,25 +344,6 @@ class MaxRetries(_PositiveInt):
     ...
 
 
-class FlexPercent(ValueObject):
-    def __init__(self, value: float, /):
-        if value is None:
-            raise ValueError(
-                f"{self.__class__.__name__} value is required, but got None."
-            )
-        elif isinstance(value, float):
-            if value < 0 or value > 1:
-                raise ValueError(
-                    f"{self.__class__.__name__} value must be between 0 and 1, but got {value!r}."
-                )
-        else:
-            raise TypeError(
-                f"{self.__class__.__name__} expects a float, but got {value!r}"
-            )
-
-        super().__init__(value)
-
-
 class Timestamp(ValueObject):
     def __init__(self, value: datetime.datetime, /):
         if value is None:
